Package: profvis
Title: Interactive Visualizations for Profiling R Code
Version: 0.4.0
Authors@R: c(
    person("Hadley", "Wickham", , "hadley@posit.co", role = c("aut", "cre")),
    person("Winston", "Chang", role = "aut"),
    person("Javier", "Luraschi", role = "aut"),
    person("Timothy", "Mastny", role = "aut"),
    person("Posit Software, PBC", role = c("cph", "fnd")),
    person(, "jQuery Foundation", role = "cph",
           comment = "jQuery library"),
    person(, "jQuery contributors", role = c("ctb", "cph"),
           comment = "jQuery library; authors listed in inst/htmlwidgets/lib/jquery/AUTHORS.txt"),
    person("Mike", "Bostock", role = c("ctb", "cph"),
           comment = "D3 library"),
    person(, "D3 contributors", role = "ctb",
           comment = "D3 library"),
    person("Ivan", "Sagalaev", role = c("ctb", "cph"),
           comment = "highlight.js library")
  )
Description: Interactive visualizations for profiling R code.
License: MIT + file LICENSE
URL: https://profvis.r-lib.org, https://github.com/r-lib/profvis
BugReports: https://github.com/r-lib/profvis/issues
Depends: R (>= 4.0)
Imports: htmlwidgets (>= 0.3.2), rlang (>= 1.1.0), vctrs
Suggests: htmltools, knitr, rmarkdown, shiny, testthat (>= 3.0.0)
VignetteBuilder: knitr
Config/Needs/website: tidyverse/tidytemplate, rmarkdown
Config/testthat/edition: 3
Encoding: UTF-8
RoxygenNote: 7.3.2
NeedsCompilation: yes
Packaged: 2024-09-19 19:26:06 UTC; hadleywickham
Author: Hadley Wickham [aut, cre],
  Winston Chang [aut],
  Javier Luraschi [aut],
  Timothy Mastny [aut],
  Posit Software, PBC [cph, fnd],
  jQuery Foundation [cph] (jQuery library),
  jQuery contributors [ctb, cph] (jQuery library; authors listed in
    inst/htmlwidgets/lib/jquery/AUTHORS.txt),
  Mike Bostock [ctb, cph] (D3 library),
  D3 contributors [ctb] (D3 library),
  Ivan Sagalaev [ctb, cph] (highlight.js library)
Maintainer: Hadley Wickham <hadley@posit.co>
Repository: CRAN
Date/Publication: 2024-09-20 11:50:10 UTC
Built: R 4.6.0; aarch64-apple-darwin23; 2026-04-26 15:55:11 UTC; unix
Archs: profvis.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method(print,profvis)
export(parse_rprof)
export(pause)
export(profvis)
export(profvisOutput)
export(profvis_server)
export(profvis_ui)
export(renderProfvis)
import(htmlwidgets)
import(rlang)
importFrom(utils,Rprof)
useDynLib(profvis, .registration = TRUE, .fixes = "c_")

profvis
=======

<!-- badges: start -->
[![R-CMD-check](https://github.com/r-lib/profvis/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/r-lib/profvis/actions/workflows/R-CMD-check.yaml)
[![Codecov test coverage](https://codecov.io/gh/r-lib/profvis/graph/badge.svg)](https://app.codecov.io/gh/r-lib/profvis)
<!-- badges: end -->

profvis is a tool for visualizing code profiling data from R. It creates a web page which provides a graphical interface for exploring the data.


## Installation

```R
install.packages("profvis")
```

## Example

To run code with profiling, wrap the expression in `profvis()`. By default, this will result in the interactive profile visualizer opening in a web browser.

```R
library(profvis)

f <- function() {
  pause(0.1)
  g()
  h()
}
g <- function() {
  pause(0.1)
  h()
}
h <- function() {
  pause(0.1)
}

profvis(f())
```


The `profvis()` call returns an [htmlwidget](http://www.htmlwidgets.org/), which by default when printed opens a web browser. If you wish to save the object, it won't open the browser at first, but you can view it later by typing the variable name at the console, or calling `print()` on it.

```R
p <- profvis(f())

# View it with:
p
# or print(p)
```

# These tests must be run manually at the console because R generates source
# refs differently at the console than it does in a test file that is
# `source()`ed.
#
# Unless otherwise specified, the test is simply to run the code and see if
# it produces a usable profile.

if (interactive()) {

  # Files that are sourced. This should show source for <expr> and source.R.
  source("tests/manual-test-source.R", keep.source = TRUE)
  profvis({ f() })


  # Functions with spaces
  profvis({
    `f f` <- function() { pause(0.1) }
    `f f`()
    gc()
  })


  # Missing exprs. The function `f` will not have its source displayed. Also,
  # this should not error.
  f <- function() {

    pause(0.1)
  }
  profvis({ f() })


  # Directories in source file path are not exactly the same as the package
  # name. For example, a file name might be "tidyverse-ggplot2-9f7b08c/R/plot.r"
  devtools::install_github("tidyverse/ggplot2", args = "--with-keep.source",
    force = TRUE)
  library(ggplot2)
  profvis({
    g <- ggplot(mtcars, aes(wt, mpg)) + geom_point()
    print(g)
  })


  # Should show source for files loaded with load_all()
  devtools::load_all("../ggplot2")
  profvis({
    g <- ggplot(mtcars, aes(wt, mpg)) + geom_point()
    print(g)
  })

}

## -----------------------------------------------------------------------------
library(profvis)
library(knitr)

knitr::opts_chunk$set(collapse = TRUE, comment = "#>")

# Make output a little less tall by default
registerS3method("knit_print", "htmlwidget", function(x, ...) {
  # Get the chunk height
  height <- knitr::opts_current$get("height")
  if (length(height) > 0 && height != FALSE)
    x$height <- height
  else
    x$height <- "450px"

  htmlwidgets:::knit_print.htmlwidget(x, ...)
})

## ----abline, fig.show="hide"--------------------------------------------------
library(profvis)

profvis({
  df <- data.frame(x = rnorm(5e5), y = rnorm(5e5))

  plot(y ~ x, data = df)
  m <- lm(y ~ x, data = df)
  abline(m, col = "red")
})

#' Generate profiler lines for an expression
#'
#' @param expr Expression to profile. The expression is repeatedly
#'   evaluated until `Rprof()` produces an output. Can _be_ a quosure
#'   injected with [rlang::inject()] but cannot _contain_ injected
#'   quosures.
#' @param ... Arguments passed to `Rprof()`.
#' @param trim_stack Whether to trim the current call stack from the
#'   profiles.
#' @param rerun Regexp or `NULL`. If supplied, resamples a new
#'   profile until the regexp matches the modal profile
#'   stack. Metadata is removed from the profiles before matching and
#'   taking the modal value.
#' @noRd
rprof_lines <- function(expr,
                        ...,
                        interval = 0.010,
                        filter.callframes = FALSE,
                        trim_stack = TRUE,
                        rerun = FALSE) {
  c(expr, env) %<-% enquo0_list(expr)

  lines <- character()

  prof_file <- tempfile("profvis-snapshot", fileext = ".prof")
  on.exit(unlink(prof_file), add = TRUE)

  if (has_simplify()) {
    args <- list(filter.callframes = filter.callframes)
  } else {
    args <- NULL
  }
  on.exit(Rprof(NULL), add = TRUE)

  while (TRUE) {
    env_bind_lazy(current_env(), do = !!expr, .eval_env = env)

    gc()
    inject(Rprof(
      prof_file,
      ...,
      interval = interval,
      !!!args
    ))

    do
    Rprof(NULL)

    lines <- zap_header(readLines(prof_file, warn = FALSE))

    if (trim_stack) {
      suffix <- rprof_current_suffix(env, filter.callframes, ...)
      lines <- gsub(suffix, "", lines)
    }

    if (prof_matches(lines, rerun)) {
      break
    }
  }

  lines
}

re_srcref <- "\\d+#\\d+"
re_srcref_opt <- sprintf(" (%s )?", re_srcref)

rprof_current_suffix <- function(env, simplify, ...) {
  if (simplify && getRversion() >= "4.0.3") {
    # We need to call the suffix routine from the caller frame. We
    # inline a closure in the call so we can refer to here despite
    # evaluating in a foreign environment. Evaluation is done through
    # a promise to keep the call stack simple.
    call <- call2(function() rprof_current_suffix_linear(...))
    env_bind_lazy(current_env(), do = !!call, .eval_env = env)
    do
  } else {
    rprof_current_suffix_full(...)
  }
}
rprof_current_suffix_full <- function(...) {
  lines <- rprof_lines(
    pause(RPROF_SUFFIX_PAUSE_TIME),
    trim_stack = FALSE,
    ...,
    rerun = "rprof_current_suffix_full"
  )
  line <- modal_value(zap_meta_data(lines))

  pattern <- sprintf(" \"rprof_current_suffix\"( %s)?", re_srcref)
  pos <- gregexpr(pattern, line)[[1]]

  if (length(pos) != 1 || pos < 0) {
    stop("Unexpected state in `rprof_current_suffix()`.")
  }
  suffix <- substring(line, pos + attr(pos, "match.length"))

  suffix <- gsub_srcref_as_wildcards(suffix)
  paste0(suffix, "$")
}

rprof_current_suffix_linear <- function(..., filter.callframes = NULL) {
  lines <- rprof_lines(
    pause(RPROF_SUFFIX_PAUSE_TIME),
    trim_stack = FALSE,
    ...,
    filter.callframes = TRUE,
    rerun = "rprof_current_suffix_linear"
  )
  line <- modal_value(zap_meta_data(lines))

  pattern <- sprintf(
    "^\"pause\"%s\"rprof_current_suffix_linear\"%s\"<Anonymous>\"%s",
    re_srcref_opt,
    re_srcref_opt,
    re_srcref_opt
  )
  suffix <- sub(pattern, "", line)

  suffix <- gsub_srcref_as_wildcards(suffix)
  paste0(suffix, "$")
}

RPROF_SUFFIX_PAUSE_TIME <- 0.100

# File labels of the suffix will differ with those of the actual
# profiles
gsub_srcref_as_wildcards <- function(lines) {
  # Strip all existing srcrefs
  lines <- gsub("\\d+#\\d+ ", "", lines)

  # Add wildcards for srcrefs
  lines <- gsub("\" \"", sprintf("\"%s\"", re_srcref_opt), lines, fixed = TRUE)

  lines
}

utils::globalVariables("do")

has_simplify <- function() {
  getRversion() >= "4.0.3"
}

skip_on_cran_if_not_ci()

test_that("`rprof_lines()` collects profiles", {
  f <- function() pause(TEST_PAUSE_TIME)

  out <- rprof_lines(f(), rerun = "pause")
  expect_snapshot(writeLines(modal_value0(out)))

  expect_snapshot0(cat_rprof(f()))
})

test_that("`filter.callframes` filters out intervening frames", {
  skip_if_not(has_simplify())

  # Chains of calls are kept
  f <- function() g()
  g <- function() h()
  h <- function() pause(TEST_PAUSE_TIME)
  expect_snapshot0(cat_rprof(f(), filter.callframes = TRUE))

  # Intervening frames are discarded
  f <- function() identity(identity(pause(TEST_PAUSE_TIME)))
  expect_snapshot0(cat_rprof(f(), filter.callframes = TRUE))
})

test_that("stack is correctly stripped even with metadata profiling", {
  f <- function() pause(TEST_PAUSE_TIME)
  zap <- function(lines) modal_value0(zap_trailing_space(zap_srcref(zap_meta_data(lines))))

  metadata <- rprof_lines(
    f(),
    line.profiling = TRUE,
    memory.profiling = TRUE,
    filter.callframes = FALSE,
    rerun = "pause"
  )
  expect_snapshot(writeLines(zap(metadata)))

  metadata_simplified <- rprof_lines(
    f(),
    line.profiling = TRUE,
    memory.profiling = TRUE,
    filter.callframes = TRUE,
    rerun = "pause"
  )
  expect_snapshot(writeLines(zap(metadata_simplified)))
})

test_that("`pause()` does not include .Call() when `line.profiling` is set", {
  f <- function() pause(TEST_PAUSE_TIME)

  # `pause()` should appear first on the line
  out <- modal_value(rprof_lines(f(), line.profiling = TRUE, rerun = "pause"))
  expect_true(grepl("^\"pause\" ", out))
})

test_that("srcrefs do not prevent suffix replacement", {
  line <- ":1509169:3184799:91929040:0:\"pause\" 1#1 \"f\" \"doTryCatch\" \"tryCatchOne\" \"tryCatchList\" \"doTryCatch\" \"tryCatchOne\" \"tryCatchList\" \"tryCatch\" 2#193 \"with_profvis_handlers\" 2#151 \"profvis\" "
  suffix <- "\"doTryCatch\" \"tryCatchOne\" \"tryCatchList\" \"doTryCatch\" \"tryCatchOne\" \"tryCatchList\" \"tryCatch\" 2#193 \"with_profvis_handlers\" 2#151 \"profvis\" $"
  re <- gsub_srcref_as_wildcards(suffix)
  expect_equal(
    gsub(re, "", line),
    ":1509169:3184799:91929040:0:\"pause\" 1#1 \"f\" "
  )
})

TEST_PAUSE_TIME <- 0.050

call_stacks <- function(x) {
  prof <- x$x$message$prof
  stacks <- split(prof$label, prof$time)
  vapply(stacks, paste, "", collapse = " ")
}

modal_call <- function(x) {
  modal_value0(call_stacks(x))
}

profile_calls <- function(x) {
  prof <- x$x$message$prof
  stacks <- split(prof$label, prof$time)
  vapply(stacks, paste, "", collapse = " ")
}

profile_mode <- function(x) {
  modal_value0(profile_calls(x))
}

# Drop NULLs from a list
drop_nulls <- function(x) {
  x[!vapply(x, is.null, logical(1))]
}

# Everything above this function in the stack will be hidden by default in the
# flamegraph.
..stacktraceoff.. <- function(x) x

modal_value0 <- function(x) {
  if (!length(x)) {
    return(NULL)
  }

  if (is.list(x)) {
    abort("Can't use `modal_value()` with lists.")
  }
  self_split <- unname(split(x, x))

  lens <- lengths(self_split)
  max_locs <- which(lens == max(lens))

  if (length(max_locs) != 1) {
    return(NULL)
  }

  modal <- self_split[[max_locs]]
  modal[[1]]
}
modal_value <- function(x) {
  modal_value0(x) %||% abort("Expected modal value.")
}

enquo0_list <- function(arg) {
  quo <- inject(enquo0(!!substitute(arg)), caller_env())

  # Warn if there are any embedded quosures as these are not supported
  quo_squash(quo, warn = TRUE)

  list(
    expr = quo_get_expr(quo),
    env = quo_get_env(quo)
  )
}


split_in_half <- function(x, pattern, fixed = FALSE, perl = FALSE) {
  pos <- regexpr(pattern, x, fixed = fixed, perl = perl)
  
  start <- as.vector(pos) - 1
  length <- attr(pos, "match.length")
  
  no_match <- !is.na(pos) & pos == -1L
  length[no_match] <- 0
  start[no_match] <- nchar(x)[no_match]

  cbind(
    substr(x, 1, start),
    substr(x, start + length + 1, nchar(x))
  )
}

test_that("split_in_half handles common cases", {
  out <- split_in_half(c("a-b", "a---b", "ab", "", NA), "-+", perl = TRUE)
  expect_equal(out, cbind(
    c("a", "a", "ab", "", NA),
    c("b", "b", "", "", NA)
  ))
})

#' profvis UI for Shiny Apps
#'
#' Use this Shiny module to inject profvis controls into your Shiny app. The
#' profvis Shiny module injects UI that can be used to start and stop profiling,
#' and either view the results in the profvis UI or download the raw .Rprof
#' data. It is highly recommended that this be used for testing and debugging
#' only, and not included in production apps!
#'
#' The usual way to use profvis with Shiny is to simply call
#' `profvis(shiny::runApp())`, but this may not always be possible or desirable:
#' first, if you only want to profile a particular interaction in the Shiny app
#' and not capture all the calculations involved in starting up the app and
#' getting it into the correct state; and second, if you're trying to profile an
#' application that's been deployed to a server.
#'
#' For more details on how to invoke Shiny modules, see [this
#' article](https://shiny.rstudio.com/articles/modules.html).
#'
#' @param id Output id from `profvis_server`.
#'
#' @examples
#' # In order to avoid "Hit <Return> to see next plot" prompts,
#' # run this example with `example(profvis_ui, ask=FALSE)`
#'
#' if(interactive()) {
#'   library(shiny)
#'   shinyApp(
#'     fluidPage(
#'       plotOutput("plot"),
#'       actionButton("new", "New plot"),
#'       profvis_ui("profiler")
#'     ),
#'     function(input, output, session) {
#'       callModule(profvis_server, "profiler")
#'
#'       output$plot <- renderPlot({
#'         input$new
#'         boxplot(mpg ~ cyl, data = mtcars)
#'       })
#'     }
#'   )
#' }
#'
#' @export
profvis_ui <- function(id) {
  if (!requireNamespace("shiny", quietly = TRUE)) {
    stop('profvis_ui requires the shiny package.')
  }
  ns <- shiny::NS(id)

  style <- htmltools::css(
    padding = "6px",
    white_space = "nowrap",
    top = "-1px",
    border_top_left_radius = "0",
    border_top_right_radius = "0",
    box_shadow = "none",
    z_index = 9000
  )

  shiny::tagList(
    htmltools::tags$style(
      ".profvis-module-container:empty() { visibility: hidden; }"
    ),
    shiny::fixedPanel(
      top = 0, left = -1, width = "auto", height = "auto",
      class = "profvis-module-container well", style = style, draggable = TRUE,

      shiny::uiOutput(ns("button_group"), class = "btn-group")
    ),
    # TODO: Make htmlDependency
    shiny::singleton(shiny::includeScript(system.file("shinymodule/draggable-helper.js", package = "profvis")))
  )
}

#' @param input,output,session Arguments provided by
#'   [shiny::callModule()].
#' @param dir Output directory to save Rprof files.
#'
#' @rdname profvis_ui
#' @export
profvis_server <- function(input, output, session, dir = ".") {
  if (!requireNamespace("shiny", quietly = TRUE)) {
    stop('profvis_server requires the shiny package.')
  }
  # Whether we're currently profiling
  profiling <- shiny::reactiveVal(FALSE)
  # The current/most recent profile
  current_profile <- shiny::reactiveVal(NULL)

  profiles <- function() {
    dir(dir, pattern = "\\.Rprof$")
  }

  shiny::setBookmarkExclude(c("start_rprof", "browse", "dl_rprof", "dl_profvis", "download"))

  shiny::observeEvent(input$start_rprof, {
    if (!profiling()) {
      proffile <- file.path(dir, strftime(Sys.time(), "%Y-%m-%d_%H-%M-%S.Rprof"))
      Rprof(proffile,
        interval = 0.01, line.profiling = TRUE,
        gc.profiling = TRUE, memory.profiling = TRUE)
      current_profile(proffile)
      profiling(TRUE)
    }
  })

  output$button_group <- shiny::renderUI({
    profiling()

    ns <- session$ns

    shiny::isolate({
      browseBtn <- shiny::actionButton(class = "btn-xs", ns("browse"), NULL, shiny::icon("list-ul"))

      if (!profiling()) {
        htmltools::tagList(
          shiny::actionButton(class = "btn-xs", ns("start_rprof"), "Start profiling", shiny::icon("play")),
          if (length(profiles()) > 0) browseBtn
        )
      } else {
        # Register a URL for the "Stop Recording" button to go to.
        # Requesting this URL will stop the current profiling session, update
        # the profiling() reactiveVal, and return a new profvis.
        url <- session$registerDataObj("stop_profvis_module", list(), function(data, req) {
          shiny::isolate({
            Rprof(NULL)
            profiling(FALSE)

            # profiling(FALSE) should cause a flushReact, but doesn't. This
            # invalidateLater is a hack to force one (since it's inside an
            # isolate, it otherwise has no effect).
            shiny::invalidateLater(50)

            if (is.null(current_profile())) {
              stop("Invalid state detected")
            }

            # Create a profvis and save it to a self-contained temp .html file
            p <- profvis(prof_input = current_profile())
            outfile <- tempfile("profvis", fileext = ".html")
            htmlwidgets::saveWidget(p, outfile)

            # Return as HTML. Since owned=TRUE, httpuv will take care of deleting
            # the temp file when it's done streaming it to the client.
            list(
              status = 200L,
              headers = list(
                "Content-Type" = "text/html;charset=utf-8"
              ),
              body = list(
                file = outfile,
                owned = TRUE
              )
            )
          })
        })

        htmltools::tagList(
          htmltools::tags$a(class = "btn btn-default btn-xs", target = "_blank", href = url, shiny::icon("stop"), "Stop profiling")
        )
      }
    })
  })

  shiny::observeEvent(input$browse, {
    ns <- session$ns

    shiny::showModal(shiny::modalDialog(
      shiny::uiOutput(ns("download_select")),
      shiny::downloadButton(ns("dl_rprof"), "Download as Rprof", class = "btn-xs"),
      shiny::downloadButton(ns("dl_profvis"), "Download as profvis", class = "btn-xs")
    ))
  })

  shiny::onSessionEnded(function() {
    # Make sure we stop profiling when session exits
    Rprof(NULL)
  })

  output$download_select <- shiny::renderUI({
    shiny::req(!profiling(), cancelOutput = TRUE)
    ns <- session$ns
    shiny::selectInput(ns("download"), "Select profile to download",
                choices = sort(profiles(), decreasing = TRUE)
    )
  })

  # Validate input$download so we don't just let the user download whatever
  # file they want from the server.
  download <- shiny::reactive({
    dl <- input$download
    shiny::validate(shiny::need(isTRUE(dl %in% profiles()), "Illegal download or not found"))
    dl
  })

  output$dl_rprof <- shiny::downloadHandler(
    filename = function() {
      file.path(dir, download())
    },
    content = function(file) {
      file.copy(
        file.path(dir, download()),
        file
      )
    }
  )

  output$dl_profvis <- shiny::downloadHandler(
    filename = function() {
      file.path(dir, sub("Rprof$", "html", download()))
    },
    content = function(file) {
      p <- profvis(prof_input = download())
      htmlwidgets::saveWidget(p, file)
    }
  )
}

#' Parse Rprof output file for use with profvis
#'
#' @param path Path to the [Rprof()] output file.
#' @param expr_source If any source refs in the profiling output have an empty
#'   filename, that means they refer to code executed at the R console. This
#'   code can be captured and passed (as a string) as the `expr_source`
#'   argument.
#' @keywords internal
#' @export
parse_rprof <- function(path = "Rprof.out", expr_source = NULL) {
  parse_rprof_lines(readLines(path), expr_source = expr_source)
}
parse_rprof_lines <- function(lines, expr_source = NULL) {
  stopifnot(is_character(lines))

  if (length(lines) < 2) {
    stop("No parsing data available. Maybe your function was too fast?")
  }

  # Parse header, including interval (in ms)
  opts <- strsplit(lines[[1]], ": ", fixed = TRUE)[[1]]
  interval <- as.numeric(strsplit(opts[length(opts)], "=", fixed = TRUE)[[1]][2]) / 1e3
  lines <- lines[-1]

  # Separate file labels and profiling data
  is_label <- grepl("^#", lines)

  label_lines <- lines[is_label]
  label_pieces <- split_in_half(label_lines, ": ", fixed = TRUE)
  labels <- data.frame(
    label = as.numeric(sub("^#File ", "", label_pieces[, 1])),
    path = label_pieces[, 2],
    stringsAsFactors = FALSE
  )

  # Parse profiling data -----------------
  prof_lines <- lines[!is_label]

  prof_data <- sub(' +$', '', prof_lines)

  # Memory profiles start with ':'
  has_memory <- length(prof_data) > 0 && substr(prof_data[[1]], 1, 1) == ":"

  # Extract memory data from ':m1:m2:m3:d:"c1" "c2" "c3"', and remove the memory
  # stuff from the prof_data strings.
  if (has_memory) {
    mem_data <- gsub("^:(\\d+:\\d+:\\d+:\\d+):.*", "\\1", prof_data)
    mem_data <- strsplit(mem_data, ":", fixed = TRUE)
    prof_data <- zap_mem_prefix(prof_data)
  } else {
    mem_data <- rep(NA_character_, length(prof_data))
  }

  # Convert frames with srcrefs from:
  #  "foo" 2#8
  # to
  #  "foo",2#8
  prof_data <- gsub('" (\\d+#\\d+)', '",\\1', prof_data)
  # But if the line starts with a <GC>, it shouldn't be joined like that.
  # Convert:
  #  <GC>,1#7 "foo"
  # back to
  #  <GC> 1#7 "foo"
  prof_data <- gsub('^"<GC>",', '"<GC>" ', prof_data)

  # # Split by ' ' for call stack
  # prof_data <- strsplit(prof_data, " ")
  #
  # prof_data <- lapply(prof_data, function(s) {
  #   if (identical(s, "")) character(0)
  #   else s
  # })

  # Parse each line into a separate data frame
  prof_data <- mapply(prof_data, mem_data, seq_along(prof_data), FUN = function(line, mem, time) {
    memalloc <- 0
    if (has_memory) {
      # See memory allocation on r-sources (memory.c):
      # https://github.com/wch/r-source/blob/tags/R-3-0-0/src/main/memory.c#L1845
      # Memory is defined as: small:big:nodes:dupes. Originally, we tracked
      # mem[1:3] to include 'nodes' which track expression allocations.
      # However, the 3rd parameter is internal to the R execution engine since
      # it tracks all expression references and can yield information that's
      # confusing to users. For instance, profiling profvis::pause(1) can yield
      # several hundred MB due to busy waits of pause that trigger significant
      # creation of expressions that is not enterily useful to the end user.
      memalloc <- sum(as.numeric(mem[1:2])) / 1024 ^ 2

      # get_current_mem provides the results as either R_SmallVallocSize or R_LargeVallocSize
      # which are internal untis of allocation.
      # https://github.com/wch/r-source/blob/tags/R-3-0-0/src/main/memory.c#L2291.
      #
      # R_SmallVallocSize maps to alloc_size; alloc_size is assigned from size, which depending on
      # the type gets calculated with a macro, for instance, using FLOAT2VEC.
      # https://github.com/wch/r-source/blob/tags/R-3-0-0/src/main/memory.c#L2374
      #
      # FLOAT2VEC and similar functions always divide by sizeof(VECREC).
      # https://github.com/wch/r-source/blob/tags/R-3-0-0/src/include/Defn.h#L400
      #
      # VECREC is defined as follows:
      # typedef struct {
      #   union {
      #     SEXP     backpointer;
      #     double   align;
      #   } u;
      # } VECREC, *VECP;
      #
      # SEXP is defined as typedef struct SEXPREC { ... } SEXPREC, *SEXP;
      # Therefore, SEXP being a pointer if of variable length across different platforms.
      # https://svn.r-project.org/R/trunk/src/include/Rinternals.h
      #
      # On the other hand, align is always a double of 64 bits for both, 64 and 32bit platforms.
      #
      # Therefore, this results needs to be multiplied by 8 bytes.
      memalloc <- memalloc * 8
    }

    # Replace empty strings with character(0); otherwise causes incorrect output
    # later.
    if (identical(line, ""))
      line <- character(0)

    labels <- scan(text = line, what = character(0), quiet = TRUE)

    # If an element in `labels` is just a bare srcref without label, it doesn't
    # actually refer to a function call on the call stack -- instead, it just
    # means that the line of code is being evaluated. This can happen in either
    # of the first 2 elements in `labels`, because it could be "3#19", or it
    # could be "<GC> 3#19" -- the <GC> doesn't count as a real label.
    #
    # Note how the first lineprof() call differs from the ones in the loop:
    # https://github.com/wch/r-source/blob/be7197f/src/main/eval.c#L228-L244 In
    # this case, we'll use NA as the label for now, and later insert the line of
    # source code.
    bare_srcref_idx <- grep("^\\d+#\\d+$", labels[1:2])

    # If found the bare srcref, insert an NA before it.
    if (length(bare_srcref_idx) > 0) {
      after_idx <- seq.int(bare_srcref_idx, length(labels))
      labels <- c(labels[-after_idx], NA_character_, labels[after_idx])
    }

    # Extract the srcrefs. These have the form ",3#12", or "3#12" if it was the
    # first item on the line.
    ref_idx <- grep('^,?\\d+#\\d+$', labels)

    # The number of calls on the stack
    n_calls <- length(labels) - length(ref_idx)

    # Create char vector with srcref strings, of form "3#12" or ",3#12".
    ref_strs <- rep(NA_character_, n_calls)
    ref_strs[ref_idx - seq_along(ref_idx)] <- labels[ref_idx]

    # Remove srcref text from `labels`. Make sure length is >0 because if length
    # is 0, labels[-integer(0)] will drop all entries.
    if (length(ref_idx) > 0)
      labels <- labels[-ref_idx]

    # Get file and line numbers
    ref_strs <- sub('^,', '', ref_strs)
    filenum <- as.numeric(sub('#.*', '', ref_strs))
    linenum <- as.numeric(sub('.*#', '', ref_strs))

    nrows <- length(labels)
    # Return what is essentially a data frame, but in list format because R is
    # slow at creating data frames here, and slow at rbinding them later. Doing
    # it with lists is about 4-5x faster than with data frames.
    list(
      time = rep(time, nrows),
      depth = if (nrows == 0) integer(0) else seq(nrows, 1),
      label = labels,
      filenum = filenum,
      linenum = linenum,
      # Using numeric(0) for memalloc can be slightly erroneous because memory
      # could have been allocated here due to stuff that happened in the part of
      # the stack that got trimmed off earlier. But there's another way to
      # represent memory usage because it's not associated with a line or
      # function label, only a time stamp, and profvis doesn't record memory
      # usage by time alone -- it must be associated with a function call and
      # optionally, a line of code.
      memalloc = rep(memalloc, nrows)
    )
  }, SIMPLIFY = FALSE, USE.NAMES = FALSE)


  extract_vector <- function(x, name) {
    vecs <- lapply(x, `[[`, name)
    do.call(c, vecs)
  }

  # Bind all the pseudo data-frames together, into a real data frame.
  prof_data <- data.frame(
    time = extract_vector(prof_data, "time"),
    depth = extract_vector(prof_data, "depth"),
    label = extract_vector(prof_data, "label"),
    filenum = extract_vector(prof_data, "filenum"),
    linenum = extract_vector(prof_data, "linenum"),
    memalloc = extract_vector(prof_data, "memalloc"),
    stringsAsFactors = FALSE
  )

  # Compute memory changes
  prof_data$meminc <- append(0, diff(prof_data$memalloc))

  # Add filenames
  prof_data$filename <- labels$path[prof_data$filenum]

  # Get code file contents ---------------------------
  filenames <- unique(prof_data$filename)
  # Drop NA
  filenames <- filenames[!is.na(filenames)]


  file_contents <- get_file_contents(filenames, expr_source)

  # Trim filenames to make output a bit easier to interpret
  prof_data$filename <- trim_filenames(prof_data$filename)
  normpaths <- normalizePath(names(file_contents), winslash = "/", mustWork = FALSE)
  # Workaround for different behavior of normalizePath on Windows. Need to convert
  # "C:/path/to/file/<expr>" back to just "<expr>".
  if (.Platform$OS.type == "windows") {
    normpaths <- sub(file.path(getwd(), "<expr>"), "<expr>", normpaths, fixed = TRUE)
  }
  names(file_contents) <- trim_filenames(names(file_contents))

  # Remove srcref info from the prof_data in cases where no file is present.
  no_file_idx <- !(prof_data$filename %in% names(file_contents))
  prof_data$filename[no_file_idx] <- NA
  prof_data$filenum[no_file_idx] <- NA
  prof_data$linenum[no_file_idx] <- NA

  # Because we removed srcrefs when no file is present, there can be cases where
  # the label is NA and we couldn't read the file. This is when the profiler
  # output is like '1#2 "foo" "bar"' -- when the first item is a ref that
  # points to a file we couldn't read. We need to remove these NAs because we
  # don't have any useful information about them.
  prof_data <- prof_data[!(is.na(prof_data$label) & no_file_idx), ]

  # Add labels for where there's a srcref but no function on the call stack.
  # This can happen for frames at the top level.
  prof_data <- insert_code_line_labels(prof_data, file_contents)

  # Convert file_contents to a format suitable for client
  file_contents <- mapply(names(file_contents), file_contents, normpaths,
    FUN = function(filename, content, normpath) {
      list(filename = filename, content = content, normpath = normpath)
    }, SIMPLIFY = FALSE, USE.NAMES = FALSE)

  list(
    prof = prof_data,
    interval = interval,
    files = file_contents
  )
}

zap_mem_prefix <- function(lines) {
  gsub("^:\\d+:\\d+:\\d+:\\d+:", "\\1", lines)
}
zap_file_labels <- function(lines) {
  lines[!grepl("^#", lines)]
}
zap_srcref <- function(lines) {
  gsub(" \\d+#\\d+", "", lines)
}
zap_meta_data <- function(lines) {
  lines <- zap_file_labels(lines)
  lines <- zap_mem_prefix(lines)
  lines
}
zap_header <- function(lines) {
  lines <- lines[-1]
  lines <- zap_file_labels(lines)
  lines
}

# For any rows where label is NA and there's a srcref, insert the line of code
# as the label.
insert_code_line_labels <- function(prof_data, file_contents) {
  file_label_contents <- lapply(file_contents, function(content) {
    content <- strsplit(content, "\n", fixed = TRUE)[[1]]
    sub("^ +", "", content)
  })

  # Indices where a filename is present and the label is NA
  filename_idx <- !is.na(prof_data$filename) & is.na(prof_data$label)

  # Get the labels
  labels <- mapply(
    prof_data$filename[filename_idx],
    prof_data$linenum[filename_idx],
    FUN = function(filename, linenum) {
      if (filename == "")
        return("")
      file_label_contents[[filename]][linenum]
    }, SIMPLIFY = FALSE)
  labels <- unlist(labels, use.names = FALSE)
  # Insert the labels at appropriate indices
  prof_data$label[filename_idx] <- labels

  prof_data
}


trim_filenames <- function(filenames) {
  # Strip off current working directory from filenames
  filenames <- sub(getwd(), "", filenames, fixed = TRUE)

  # Replace /xxx/yyy/package/R/zzz.R with package/R/zzz.R, and same for inst/.
  filenames <- sub("^.*?([^/]+/(R|inst)/.*\\.R$)", "\\1", filenames, ignore.case = TRUE)

  filenames
}

# The profile data is sorted by time by default. To sort it
# alphabetically we create a data frame of stacks where columns
# represent stack depth and rows represent samples.
# `vctrs::vec_order()` then gives us the sorting key we need.
prof_sort <- function(prof) {
  # Split profile data frame by `time`. Each split corresponds to a
  # single line of the original Rprof output, i.e. a single sampled
  # stack.
  prof_split <- vctrs::vec_split(prof, prof$time)$val

  max_depth <- max(vapply(prof_split, nrow, integer(1)))
  n_samples <- length(prof_split)

  # Extract labels (function names for the stack frames) and pad with NAs
  # so we can easily make a data frame
  pad <- function(x) x[seq_len(max_depth)]
  stacks <- lapply(prof_split, function(x) pad(rev(x$label)))
  stacks <- as.data.frame(do.call(rbind, stacks))

  # Reorder the profile data according to the sort key of the
  # transposed stacks
  key <- vctrs::vec_order(stacks)
  stacks <- vctrs::vec_slice(stacks, key)
  prof_split <- prof_split[key]

  # Now that stacks are in alphabetical order we sort them again by
  # contiguous run
  runs <- lapply(stacks, function(stack) {
    times <- vctrs::vec_unrep(stack)$times
    rep(rev(order(times)), times)
  })
  runs <- vctrs::data_frame(!!!runs, .name_repair = "minimal")
  prof_split <- prof_split[vctrs::vec_order(runs)]

  # Assign an increasing `time` sequence in each split
  prof_split <- Map(seq_len(n_samples), prof_split, f = function(n, split) {
    split$time <- n
    split
  })

  # Put the sorted splits back together
  vctrs::vec_rbind(!!!prof_split)
}

test_that("parsing prof files", {
  # The three lines in this file have three different kinds of call stacks.
  # They start with:
  # "test space" 1#1
  # 1#2 "test" 1#1
  # "<GC>" 1#2 "test" 1#1
  # "<GC>" "test" 1#1
  # ":"
  #
  # (The last line is empty)
  p <- parse_rprof(test_path("test-parse.prof"), expr_source = "line 1\nline 2")

  expect_identical(p$prof$time, c(1L, 2L, 2L, 3L, 3L, 3L, 4L, 4L, 5L))
  expect_identical(p$prof$depth, c(1L, 2L, 1L, 3L, 2L, 1L, 2L, 1L, 1L))
  expect_identical(p$prof$label, c("test space", "line 2", "test", "<GC>", "line 2", "test", "<GC>", "test", ":"))
  expect_identical(p$prof$filenum, c(1, 1, 1, NA, 1, 1, NA, 1, NA))
  expect_identical(p$prof$linenum, c(1, 2, 1, NA, 2, 1, NA, 1, NA))
  # Memory sizes in the test-parse.prof file were chosen to create these values
  expect_identical(p$prof$memalloc, c(136, 152, 152, 168, 168, 168, 152, 152, 152))
  expect_identical(p$prof$meminc, c(0, 16, 0, 16, 0, 0, -16, 0, 0))
  expect_identical(p$prof$filename, c("<expr>", "<expr>", "<expr>", NA, "<expr>", "<expr>", NA, "<expr>", NA))

  expect_identical(p$interval, 10)
  expect_identical(p$files,
    list(list(filename = "<expr>", content = "line 1\nline 2", normpath = "<expr>"))
  )
})

test_that("can sort profiles alphabetically (#115)", {
  prof <- eval(parse_expr(file(test_path("test-parse-sort1.R"))))

  sorted <- prof_sort(prof)

  split <- vctrs::vec_split(sorted$label, sorted$time)
  runs <- vctrs::vec_unrep(split$val)$key

  expect_equal(
    runs,
    list(
      c("pause", "foo", "f", "root"),
      c("pause", "bar", "f", "root"),
      c("pause", "foo", "root")
    )
  )

  # Regenerate `prof`
  if (FALSE) {
    root <- function() {
      for (. in 1:3) {
        f(TRUE)
        f(FALSE)
        foo()
      }
    }
    f <- function(x) if (x) foo() else bar()
    foo <- function() pause(0.05)
    bar <- function() pause(0.05)

    prof <- profvis(root())$x$message$prof
  }
})

#' @import rlang
#' @importFrom purrr map map_int map2 transpose simplify
NULL

.onLoad <- function(...) {
  run_on_load()
}

on_load <- function(expr, env = topenv(caller_env())) {
  callback <- function() eval_bare(expr, env)
  env$.__rlang_hook__. <- list2(!!!env$.__rlang_hook__., callback)
}

run_on_load <- function(env = topenv(caller_env())) {
  hook <- env$.__rlang_hook__.
  env_unbind(env, ".__rlang_hook__.")

  for (callback in hook) {
    callback()
  }

  env$.__rlang_hook__. <- NULL
}

# Script to update the version of jQuery bundled
# in projvis, at inst/htmlwidgets/lib/jquery.
# Based on simplified version of script at
# https://github.com/rstudio/shiny/blob/main/tools/updatejQuery.R

version <- "3.7.1"
extension <- ".min.js"

jq_cdn_download <- function(version, extension) {
  download.file(
    url = file.path(
      "https://code.jquery.com",
      paste0("jquery-", version, extension)
    ),
    destfile = file.path(
      "inst", "htmlwidgets", "lib", "jquery",
      paste0("jquery", extension)
    )
  )
}

jq_cdn_download(version, extension)

download.file(
  url = sprintf("https://raw.githubusercontent.com/jquery/jquery/%s/AUTHORS.txt", version),
  destfile = file.path("inst", "htmlwidgets", "lib", "jquery", "AUTHORS.txt")
)

writeLines(con = "R/version_jquery.R", c(
  "# Generated by R/updatejQuery.R; do not edit by hand",
  sprintf('version_jquery <- "%s"', version)
))

#' Pause an R process
#'
#' This function pauses an R process for some amount of time. It differs from
#' [Sys.sleep()] in that time spent in `pause` will show up in
#' profiler data. Another difference is that `pause` uses up 100\% of a CPU,
#' whereas `Sys.sleep` does not.
#'
#' @examples
#' # Wait for 0.5 seconds
#' pause(0.5)
#'
#' @param seconds Number of seconds to pause.
#' @useDynLib profvis, .registration = TRUE, .fixes = "c_"
#' @export
pause <- function(seconds) {
  if (is.integer(seconds)) {
    seconds <- as.numeric(seconds)
  }
  .Call(c_profvis_pause, seconds)
}

# This guarantees that (1) `pause()` is always compiled, even on
# `load_all()` and (2) it doesn't include source references. This in
# turn ensures consistent profile output: if the function is not
# compiled and doesn't contain srcrefs, `.Call()` is never included in
# the profiles, even when `line.profiling` is set.
on_load_pause <- function() {
  pause <<- utils::removeSource(pause)
  pause <<- compiler::cmpfun(pause)
}

test_that("pause takes expected time", {
  time <- system.time(pause(0.2))[[3]]
  # system.time is a little inaccurate so allow 10% padding
  expect_lt(abs(time - 0.2) / 0.2, 0.1)
})

test_that("works with integers", {
  expect_no_error(pause(0L))
})

test_that("pause has no srcrefs", {
  expect_equal(attr(pause, "srcref"), NULL)
})

test_that("checks its inputs", {
  expect_snapshot(error = TRUE, {
    pause(c(1, 2))
    pause("a")
  })
})

#' Profile an R expression and visualize profiling data
#'
#' This function will run an R expression with profiling, and then return an
#' htmlwidget for interactively exploring the profiling data.
#'
#' An alternate way to use `profvis` is to separately capture the profiling
#' data to a file using [Rprof()], and then pass the path to the
#' corresponding data file as the `prof_input` argument to
#' `profvis()`.
#'
#' @param expr Expression to profile. The expression will be turned into the
#'   body of a zero-argument anonymous function which is then called repeatedly
#'   as needed. This means that if you create variables inside of `expr` they
#'   will not be available outside of it.
#'
#'   The expression is repeatedly evaluated until `Rprof()` produces
#'   an output. It can _be_ a quosure injected with [rlang::inject()] but
#'   it cannot _contain_ injected quosures.
#'
#'   Not compatible with `prof_input`.
#' @param interval Interval for profiling samples, in seconds. Values less than
#'   0.005 (5 ms) will probably not result in accurate timings
#' @param prof_output Name of an Rprof output file or directory in which to save
#'   profiling data. If `NULL` (the default), a temporary file will be used
#'   and automatically removed when the function exits. For a directory, a
#'   random filename is used.
#'
#' @param prof_input The path to an [Rprof()] data file.  Not
#'   compatible with `expr` or `prof_output`.
#' @param timing The type of timing to use. Either `"elapsed"` (the
#'  default) for wall clock time, or `"cpu"` for CPU time. Wall clock time
#'  includes time spent waiting for other processes (e.g. waiting for a
#'  web page to download) so is generally more useful.
#'
#'  If `NULL`, the default, will use elapsed time where possible, i.e.
#'  on Windows or on R 4.4.0 or greater.
#' @param width Width of the htmlwidget.
#' @param height Height of the htmlwidget
#' @param split Orientation of the split bar: either `"h"` (the default) for
#'   horizontal or `"v"` for vertical.
#' @param torture Triggers garbage collection after every `torture` memory
#'   allocation call.
#'
#'   Note that memory allocation is only approximate due to the nature of the
#'   sampling profiler and garbage collection: when garbage collection triggers,
#'   memory allocations will be attributed to different lines of code. Using
#'   `torture = steps` helps prevent this, by making R trigger garbage
#'   collection after every `torture` memory allocation step.
#' @param simplify Whether to simplify the profiles by removing
#'   intervening frames caused by lazy evaluation. Equivalent to the
#'   `filter.callframes` argument to [Rprof()].
#' @param rerun If `TRUE`, `Rprof()` is run again with `expr` until a
#'   profile is actually produced. This is useful for the cases where
#'   `expr` returns too quickly, before R had time to sample a
#'   profile. Can also be a string containing a regexp to match
#'   profiles. In this case, `profvis()` reruns `expr` until the
#'   regexp matches the modal value of the profile stacks.
#'
#' @seealso [print.profvis()] for printing options.
#' @seealso [Rprof()] for more information about how the profiling
#'   data is collected.
#'
#' @examples
#' # Only run these examples in interactive R sessions
#' if (interactive()) {
#'
#' # Profile some code
#' profvis({
#'   dat <- data.frame(
#'     x = rnorm(5e4),
#'     y = rnorm(5e4)
#'   )
#'
#'   plot(x ~ y, data = dat)
#'   m <- lm(x ~ y, data = dat)
#'   abline(m, col = "red")
#' })
#'
#'
#' # Save a profile to an HTML file
#' p <- profvis({
#'   dat <- data.frame(
#'     x = rnorm(5e4),
#'     y = rnorm(5e4)
#'   )
#'
#'   plot(x ~ y, data = dat)
#'   m <- lm(x ~ y, data = dat)
#'   abline(m, col = "red")
#' })
#' htmlwidgets::saveWidget(p, "profile.html")
#'
#' # Can open in browser from R
#' browseURL("profile.html")
#'
#' }
#' @import htmlwidgets
#' @importFrom utils Rprof
#' @export
profvis <- function(expr = NULL,
                    interval = 0.01,
                    prof_output = NULL,
                    prof_input = NULL,
                    timing = NULL,
                    width = NULL,
                    height = NULL,
                    split = c("h", "v"),
                    torture = 0,
                    simplify = TRUE,
                    rerun = FALSE) {
  check_exclusive(expr, prof_input)
  split <- match.arg(split)
  c(expr_q, env) %<-% enquo0_list(expr)


  if (interval < 0.005) {
    message("Intervals smaller than ~5ms will probably not result in accurate timings.")
  }

  if (is.null(timing)) {
    if (has_event() || Sys.info()[["sysname"]] == "Windows") {
      timing <- "elapsed"
    } else {
      timing <- "cpu"
    }
  } else {
    timing <- arg_match(timing, c("elapsed", "cpu"))
  }

  if (!is.null(expr_q)) {
    # Change the srcfile to add "<expr>" as the filename. Code executed from the
    # console will have "" here, and code executed in a knitr code block will
    # have "<text>". This value is used by the profiler as the filename listed
    # in the profiler output. We need to do this to distinguish code that was
    # run in the profvis({}) code block from code that was run outside of it.
    # See https://github.com/r-lib/profvis/issues/57
    attr(expr_q, "srcfile")$filename <- "<expr>"

    # Keep original expression source code
    expr_source <- attr(expr_q, "wholeSrcref", exact = TRUE)
    expr_source <- attr(expr_source, "srcfile", exact = TRUE)$lines
    # Usually, $lines is a single string, but sometimes it can be split up into a
    # vector. Make sure it's a single string.
    expr_source <- paste(expr_source, collapse = "\n")

    prof_extension <- getOption("profvis.prof_extension", default = ".prof")

    if (is.null(prof_output) && !is.null(getOption("profvis.prof_output")))
      prof_output <- getOption("profvis.prof_output")

    remove_on_exit <- FALSE
    if (is.null(prof_output)) {
      prof_output <- tempfile(fileext = prof_extension)
      remove_on_exit <- TRUE
    }
    else {
      if (dir.exists(prof_output))
        prof_output <- tempfile(fileext = prof_extension, tmpdir = prof_output)
    }

    gc()

    if (!identical(torture, 0)) {
      gctorture2(step = torture)
      on.exit(gctorture2(step = 0), add = TRUE)
    }

    rprof_args <- drop_nulls(list(
      interval = interval,
      line.profiling = TRUE,
      gc.profiling = TRUE,
      memory.profiling = TRUE,
      event = if (has_event()) timing,
      filter.callframes = simplify
    ))

    if (remove_on_exit) {
      on.exit(unlink(prof_output), add = TRUE)
    }

    # We call the quoted expression directly inside a function to make it
    # easy to detect in both raw and simplified stack traces. The simplified
    # case is particularly tricky because evaluating a promise fails to create
    # a call on the trailing edges of the tree returned by simplification
    `__profvis_execute__` <- new_function(list(), expr_q, env)

    repeat {
      inject(Rprof(prof_output, !!!rprof_args))
      cnd <- with_profvis_handlers(`__profvis_execute__`())
      Rprof(NULL)

      lines <- readLines(prof_output)
      if (!is.null(cnd)) {
        break
      }
      if (prof_matches(zap_header(lines), rerun)) {
        break
      }
    }

    lines <- gsub('"__profvis_execute__".*$', "", lines)
  } else {
    # If we got here, we were provided a prof_input file instead of expr
    expr_source <- NULL
    prof_output <- prof_input
    lines <- readLines(prof_output)
  }

  message <- parse_rprof_lines(lines, expr_source)
  message$prof_output <- prof_output

  # Patterns to highlight on flamegraph
  message$highlight <- highlightPatterns()

  message$split <- split

  htmlwidgets::createWidget(
    name = 'profvis',
    list(message = message),
    width = width,
    height = height,
    package = 'profvis',
    sizingPolicy = htmlwidgets::sizingPolicy(
      padding = 0,
      browser.fill = TRUE,
      viewer.suppress = TRUE,
      knitr.defaultWidth = "100%",
      knitr.defaultHeight = "600px",
      knitr.figure = FALSE
    )
  )
}

prof_matches <- function(lines, rerun) {
  if (is_bool(rerun)) {
    !rerun || length(lines) > 0
  } else if (is_string(rerun)) {
    mode <- modal_value0(zap_meta_data(lines))
    !is_null(mode) && grepl(rerun, mode)
  } else {
    abort("`rerun` must be logical or a character value.")
  }
}

with_profvis_handlers <- function(expr) {
  tryCatch({
    expr
    NULL
  },
  error = function(cnd) {
    message("profvis: code exited with error:\n", cnd$message, "\n")
    cnd
  },
  interrupt = function(cnd) {
    message("profvis: interrupt received.")
    cnd
  })
}

#' Print a profvis object
#'
#' @inheritParams profvis
#' @param x The object to print.
#' @param ... Further arguments to passed on to other print methods.
#' @param aggregate If `TRUE`, the profiled stacks are aggregated by
#'   name. This makes it easier to see the big picture. Set your own
#'   global default for this argument with `options(profvis.aggregate
#'   = )`.
#' @export
print.profvis <- function(x,
                          ...,
                          width = NULL,
                          height = NULL,
                          split = NULL,
                          aggregate = NULL) {

  if (!is.null(split)) {
    split <- arg_match(split, c("h", "v"))
    x$x$message$split <- split
  }
  if (!is.null(width)) x$width <- width
  if (!is.null(height)) x$height <- height

  aggregate <- aggregate %||% getOption("profvis.aggregate") %||% FALSE
  if (aggregate) {
    x$x$message$prof <- prof_sort(x$x$message$prof)
  }

  f <- getOption("profvis.print")
  if (is.function(f)) {
    f(x, ...)
  } else {
    NextMethod()
  }
}

#' Widget output and renders functions for use in Shiny
#'
#' @param outputId Output variable for profile visualization.
#'
#' @inheritParams profvis
#' @export
profvisOutput <- function(outputId, width = '100%', height = '600px'){
  shinyWidgetOutput(outputId, 'profvis', width, height, package = 'profvis')
}

#' @param expr An expression that returns a profvis object.
#' @param env The environment in which to evaluate `expr`.
#' @param quoted Is `expr` a quoted expression (with [quote()])?
#' @export
#' @rdname profvisOutput
renderProfvis <- function(expr, env = parent.frame(), quoted = FALSE) {
  if (!quoted) { expr <- substitute(expr) } # force quoted
  shinyRenderWidget(expr, profvisOutput, env, quoted = TRUE)
}


has_event <- function() {
  getRversion() >= "4.4.0"
}

test_that("irrelevant stack trimmed from function calls (#123)", {
  skip_on_cran()
  skip_on_covr()

  f <- function() pause(TEST_PAUSE_TIME)
  g <- function() f()

  out <- profvis(g(), simplify = TRUE, rerun = "pause")
  expect_equal(profile_mode(out), "pause f g")

  out <- profvis(g(), simplify = FALSE, rerun = "pause")
  expect_equal(profile_mode(out), "pause f g")
})

test_that("irrelevant stack trimmed from inlined code (#130)", {
  skip_on_cran()
  skip_on_covr()

  out <- profvis(for (i in 1:1e4) rnorm(100), simplify = TRUE, rerun = "rnorm")
  expect_equal(profile_mode(out), "rnorm")

  out <- profvis(for (i in 1:1e4) rnorm(100), simplify = FALSE, rerun = "rnorm")
  expect_equal(profile_mode(out), "rnorm")
})

test_that("strips stack above profvis", {
  skip_on_cran()
  skip_on_covr()

  f <- function() pause(TEST_PAUSE_TIME)
  profvis_wrap <- function(...) profvis(...)

  out <- profvis_wrap(f(), simplify = TRUE, rerun = "pause")
  expect_equal(profile_mode(out), "pause f")

  out <- profvis_wrap(f(), simplify = FALSE, rerun = "pause")
  expect_equal(profile_mode(out), "pause f")
})

test_that("defaults to elapsed timing", {
  skip_on_cran()
  skip_on_covr()
  skip_if_not(has_event())

  f <- function() Sys.sleep(TEST_PAUSE_TIME)

  out <- profvis(f(), rerun = "Sys.sleep")
  expect_equal(profile_mode(out), "Sys.sleep f")
})

test_that("expr and prof_input are mutually exclusive", {
  expect_snapshot(profvis(expr = f(), prof_input = "foo.R"), error = TRUE)
})

test_that("can capture profile of code with error", {
  skip_on_cran()
  skip_on_covr()

  f <- function() {
    pause(TEST_PAUSE_TIME)
    stop("error")
  }
  expect_snapshot(out <- profvis(f(), rerun = "pause"))
  expect_equal(profile_mode(out), "pause f")
})

# A list of regexes to highlight on the flamegraph. The name of each sublist is
# the CSS class to give to cells whose labels match the regex listed. For each
# of these names, there should be a corresponding CSS class in profvis.css.
highlightPatterns <- function() {
  list(
    output = list("^output\\$"),
    gc = list("^<GC>$"),
    stacktrace = list("^\\.\\.stacktraceo(n|ff)\\.\\.$")
  )
}

get_file_contents <- function(filenames, expr_source) {
  filenames <- filenames[filenames != ""]
  names(filenames) <- filenames

  srcfile_cache <- build_srcfile_cache()
  srcfile_cache[["<expr>"]] <- expr_source

  file_contents <- lapply(filenames, function(filename) {
    fetch_cached(filename, srcfile_cache)
  })

  # If there's an <expr> entry, put it first.
  if ("<expr>" %in% names(file_contents)) {
    expr_idx <- (names(file_contents) == "<expr>")
    file_contents <- c(file_contents[expr_idx], file_contents[!expr_idx])
  }

  drop_nulls(file_contents)
}

# Fetch a file from the cache, if present. If not already present, read the file
# from disk and add it to the cache.
fetch_cached <- function(filename, srcfile_cache) {
  # If in the cache, simply return it
  if (!is.null(srcfile_cache[[filename]])) {
    return(srcfile_cache[[filename]])
  }

  # Exit if file doesn't exist locally. In some cases (e.g. a URL like
  # "http://xyz.com/" ) the `file()` call below can return a filehandle even
  # when the file is not local, and then it will error when `readChar()` is
  # called on the file. See https://github.com/r-lib/profvis/issues/73
  if (!file.exists(filename))
    return(NULL)

  # If not in the cache, try to read the file
  filehandle <- tryCatch(
    file(filename, 'rb'),
    error = function(e) NULL,
    warning = function(e) NULL
  )
  # If we can't read file, give up
  if (is.null(filehandle)) {
    return(NULL)
  }
  on.exit( close(filehandle) )

  # Add it to the cache
  srcfile_cache[[filename]] <- readChar(filename, file.info(filename)$size,
                                        useBytes = TRUE)
  srcfile_cache[[filename]]
}

build_srcfile_cache <- function(pkgs = loadedNamespaces()) {
  srcfile_cache <- new.env(parent = emptyenv())

  lapply(pkgs, function(pkg) {
    srcrefs <- get_pkg_srcrefs(pkg)
    if (length(srcrefs) > 0)
      list2env(srcrefs, srcfile_cache)
  })

  srcfile_cache
}


# Given a namespace, try to extract source code. It does this by looking at
# functions in the namespace and getting the appropriate attributes. This
# returns a named list with all sources for a package.
get_pkg_srcrefs <- function(pkg) {
  ns_env <- asNamespace(pkg)

  # Given a char vector with contents of an entire package, split out all
  # files into separate entries in a list.
  full_src_to_file_contents <- function(src) {
    # Before R 2.5.0, the first line looked like this:
    # .packageName <- "R6"
    # As of 2.5.0, that line was dropped. If that line is present, remove it.
    if (grepl("^\\.packageName <-", src[1])) {
      src <- src[-1]
    }

    # Lines which contain filenames. Have a format like:
    #   "#line 1 \"/tmp/Rtmp6W0MLC/R.INSTALL1a531f3beb59/ggplot2/R/aaa-.r\""
    filename_idx <- grep('^#line 1 "', src)
    filename_lines <- src[filename_idx]
    filenames <- sub('^#line 1 "(.*)"$', '\\1', filename_lines)

    # Starting and ending indices for the content of each file
    start_idx <- filename_idx + 1
    end_idx <- c(filename_idx[-1] - 1, length(src))

    file_contents <- mapply(start_idx, end_idx, SIMPLIFY = FALSE,
      FUN = function(start, end) {
        content <- src[seq(start, end)]
        paste(content, collapse = "\n")
      }
    )

    names(file_contents) <- filenames
    file_contents
  }

  # Get all objects in package. Need to filter out S4 mangled names (.__T__)
  ns_names <- grep("^\\.__[TC]__", ls(ns_env, all.names = TRUE), value = TRUE,
      invert = TRUE, fixed = FALSE)

  files <- list()

  for (name in ns_names) {
    x <- ns_env[[name]]
    if (is.function(x)) {
      srcref <- utils::getSrcref(x)

      # If any function lacks source refs, then no functions in the package will
      # have them. Quit early to save time.
      if (is.null(srcref))
        break

      # There are two possible formats for source refs. If the file was
      # loaded with `source()` (as with `devtools::load_all()`), the lines
      # will just be the cotents of that one file. If the file was from a
      # package that was built and installed the normal way, it will contain
      # the all sources for the entire package.
      srcfile <- attr(srcref, "srcfile", exact = TRUE)

      if (!is.null(srcfile$lines)) {
        # Was loaded with `source(). If we don't already have the source for
        # this file, save them and keep going.
        if (is.null(files[[srcfile$filename]])) {
          files[[srcfile$filename]] <- paste(srcfile$lines, collapse = "\n")
        }

      } else if (!is.null(srcfile$original$lines)) {
        # Was from a built package and therefore contains source for all
        # files in the package. We can extract source code for all files and
        # return.
        files <- full_src_to_file_contents(srcfile$original$lines)
        break

      } else {
        # Shouldn't get here -- if so, this is an unexpected configuration.
        stop("Unexpected format for source refs.")
      }
    }
  }

  files
}

---
title: "Getting started with profvis"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting started with profvis}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r}
#| include: false
library(profvis)
library(knitr)

knitr::opts_chunk$set(collapse = TRUE, comment = "#>")

# Make output a little less tall by default
registerS3method("knit_print", "htmlwidget", function(x, ...) {
  # Get the chunk height
  height <- knitr::opts_current$get("height")
  if (length(height) > 0 && height != FALSE)
    x$height <- height
  else
    x$height <- "450px"

  htmlwidgets:::knit_print.htmlwidget(x, ...)
})
```

## Introduction

[profvis](https://github.com/r-lib/profvis) is a tool for helping you to understand how R spends its time. It provides a interactive graphical interface for visualizing data from `Rprof()`, R's built-in tool for collecting profiling data.

Most R users have had times where we've wanted our code to run faster. However, it's not always clear how to accomplish this. A common approach is to rely on intuition, and on wisdom from the broader R community about speeding up R code. One drawback to this is it can lead to a focus on optimizing things that actually take a small proportion of the overall running time.

Suppose you make a loop run 5 times faster. That sounds like a huge improvement, but if that loop only takes 10% of the total time, it's still only a 8% speedup overall. Another drawback is that, although many of the commonly-held beliefs are true (for example, preallocating memory can speed things up), some are not (e.g., that `*apply` functions are inherently faster than `for` loops). This can lead us to spend time doing "optimizations" that don't really help. To make slow code faster, we need accurate information about what is making our code slow.

## Getting started

Below is an example of profvis in use. The code generates a random 50,000 row data set, draws a plot, then fits a linear model, and draws a line for the model. (The plot isn't displayed in this document, though.) If you copy and paste this code into your R console, it'll open a window with the same profvis interface that you see in this HTML document.

```{r abline, fig.show="hide"}
library(profvis)

profvis({
  df <- data.frame(x = rnorm(5e5), y = rnorm(5e5))

  plot(y ~ x, data = df)
  m <- lm(y ~ x, data = df)
  abline(m, col = "red")
})
```

On top is the code, and on the bottom is a flame graph. In the flame graph, the horizontal direction represents time in milliseconds, and the vertical direction represents the call stack. Looking at the bottom-most items on the stack, most of the time, about 2 seconds, is spent in `plot`, and then a much smaller amount of time is spent in `lm`, and almost no time at all is spent in `abline` -- it doesn't even show up on the flame graph.

Traveling up the stack, `plot` called `plot.formula`, which called `do.call`, and so on. Going up a few more levels, we can see that `plot.default` called a number of functions: first `deparse`, and later, `plot.xy`. Similarly, `lm` calls a number of different functions.

On top, we can see the amount of time spent on each line of code. This tells us, unsurprisingly, that most of the time is spent on the line with `plot`, and a little bit is spent on the line with `lm`.

The code panel also shows memory allocation and deallocation. Interpreting this information can be a little tricky, because it does not necessarily reflect memory allocated and deallcated *at* that line of code. The sampling profiler records information about memory allocations that happen between the previous sample and the current one. This means that the allocation/deallocation values on that line may have actually occurred in a previous line of code.

### Navigating profvis

profvis is interactive. You can try the following:

* As you mouse over the flame graph, information about each block will show in the info box.
* Yellow flame graph blocks have corresponding lines of code on the left. (White blocks represent code where profvis doesn't have the source code -- for example, in base R and in R packages. But see [this FAQ](#how-do-i-get-code-from-an-r-package-to-show-in-the-code-panel) if you want package code to show up in the code panel.) If you mouse over a yellow block, the corresponding line of code will be highlighted. Note that the highlighted line of code is where the yellow function is called *from*, not the content of that function.
* If you mouse over a line of code, all flame graph blocks that were called from that line will be highlighted.
* Click on a block or line of code to lock the current highlighting. Click on the background, or again on that same item to unlock the highlighting. Click on another item to lock on that item.
* Use the mouse scroll wheel or trackpad's scroll gesture to zoom in or out in the x direction.
* Click and drag on the flame graph to pan up, down, left, right.
* Double-click on the background to zoom the x axis to its original extent.
* Double-click on a flamegraph block to zoom the x axis the width of that block.

Each *block* in the flame graph represents a call to a function, or possibly multiple calls to the same function. The width of the block is proportional to the amount of time spent in that function. When a function calls another function, another block is added on top of it in the flame graph.

The profiling data has some limitations: some internal R functions don't show up in the flame graph, and it offers no insight into code that's implemented in languages other than R (e.g. C, C++, or Fortran).

### The data view

In addition to the flame graph view, profvis provides a *data view*, which can be seen by clicking on the Data tab. It provides a top-down tabular view of the profile. Click the **code** column to expand the call stack under investigation and the following columns to reason about resource allocation:

* Memory: Memory allocated or deallocated (for negative numbers) for a given call stack. This is represented in megabytes and aggregated over all the call stacks over the code in the given row.
* Time: Time spent in milliseconds. This field is also aggregated over all the call stacks executed over the code in the given row.

## How profiling data is collected

profvis uses data collected by `Rprof()`, which is part of the base R distribution. At each time interval (`profvis()` uses a default interval of 10ms), the profiler stops the R interpreter, looks at the current function call stack, and records it to a file. Because it works by *sampling*, the result isn't deterministic. Each time you profile your code, the result will be slightly different.

# `rprof_lines()` collects profiles

    Code
      writeLines(modal_value0(out))
    Output
      "pause" "f"

---

    Code
      cat_rprof(f())
    Output
      "pause" "f"

# `filter.callframes` filters out intervening frames

    Code
      cat_rprof(f(), filter.callframes = TRUE)
    Output
      "pause" "h" "g" "f" 

---

    Code
      cat_rprof(f(), filter.callframes = TRUE)
    Output
      "pause" "f" 

# stack is correctly stripped even with metadata profiling

    Code
      writeLines(zap(metadata))
    Output
      "pause" "f"

---

    Code
      writeLines(zap(metadata_simplified))
    Output
      "pause" "f"

# checks its inputs

    Code
      pause(c(1, 2))
    Condition
      Error in `pause()`:
      ! `seconds` must be a single number.
    Code
      pause("a")
    Condition
      Error in `pause()`:
      ! `seconds` must be a single number.

# expr and prof_input are mutually exclusive

    Code
      profvis(expr = f(), prof_input = "foo.R")
    Condition
      Error in `profvis()`:
      ! Exactly one of `expr` or `prof_input` must be supplied.

# can capture profile of code with error

    Code
      out <- profvis(f(), rerun = "pause")
    Message
      profvis: code exited with error:
      error
      

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pause.R
\name{pause}
\alias{pause}
\title{Pause an R process}
\usage{
pause(seconds)
}
\arguments{
\item{seconds}{Number of seconds to pause.}
}
\description{
This function pauses an R process for some amount of time. It differs from
\code{\link[=Sys.sleep]{Sys.sleep()}} in that time spent in \code{pause} will show up in
profiler data. Another difference is that \code{pause} uses up 100\\% of a CPU,
whereas \code{Sys.sleep} does not.
}
\examples{
# Wait for 0.5 seconds
pause(0.5)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/profvis.R
\name{profvis}
\alias{profvis}
\title{Profile an R expression and visualize profiling data}
\usage{
profvis(
  expr = NULL,
  interval = 0.01,
  prof_output = NULL,
  prof_input = NULL,
  timing = NULL,
  width = NULL,
  height = NULL,
  split = c("h", "v"),
  torture = 0,
  simplify = TRUE,
  rerun = FALSE
)
}
\arguments{
\item{expr}{Expression to profile. The expression will be turned into the
body of a zero-argument anonymous function which is then called repeatedly
as needed. This means that if you create variables inside of \code{expr} they
will not be available outside of it.

The expression is repeatedly evaluated until \code{Rprof()} produces
an output. It can \emph{be} a quosure injected with \code{\link[rlang:inject]{rlang::inject()}} but
it cannot \emph{contain} injected quosures.

Not compatible with \code{prof_input}.}

\item{interval}{Interval for profiling samples, in seconds. Values less than
0.005 (5 ms) will probably not result in accurate timings}

\item{prof_output}{Name of an Rprof output file or directory in which to save
profiling data. If \code{NULL} (the default), a temporary file will be used
and automatically removed when the function exits. For a directory, a
random filename is used.}

\item{prof_input}{The path to an \code{\link[=Rprof]{Rprof()}} data file.  Not
compatible with \code{expr} or \code{prof_output}.}

\item{timing}{The type of timing to use. Either \code{"elapsed"} (the
default) for wall clock time, or \code{"cpu"} for CPU time. Wall clock time
includes time spent waiting for other processes (e.g. waiting for a
web page to download) so is generally more useful.

If \code{NULL}, the default, will use elapsed time where possible, i.e.
on Windows or on R 4.4.0 or greater.}

\item{width}{Width of the htmlwidget.}

\item{height}{Height of the htmlwidget}

\item{split}{Orientation of the split bar: either \code{"h"} (the default) for
horizontal or \code{"v"} for vertical.}

\item{torture}{Triggers garbage collection after every \code{torture} memory
allocation call.

Note that memory allocation is only approximate due to the nature of the
sampling profiler and garbage collection: when garbage collection triggers,
memory allocations will be attributed to different lines of code. Using
\code{torture = steps} helps prevent this, by making R trigger garbage
collection after every \code{torture} memory allocation step.}

\item{simplify}{Whether to simplify the profiles by removing
intervening frames caused by lazy evaluation. Equivalent to the
\code{filter.callframes} argument to \code{\link[=Rprof]{Rprof()}}.}

\item{rerun}{If \code{TRUE}, \code{Rprof()} is run again with \code{expr} until a
profile is actually produced. This is useful for the cases where
\code{expr} returns too quickly, before R had time to sample a
profile. Can also be a string containing a regexp to match
profiles. In this case, \code{profvis()} reruns \code{expr} until the
regexp matches the modal value of the profile stacks.}
}
\description{
This function will run an R expression with profiling, and then return an
htmlwidget for interactively exploring the profiling data.
}
\details{
An alternate way to use \code{profvis} is to separately capture the profiling
data to a file using \code{\link[=Rprof]{Rprof()}}, and then pass the path to the
corresponding data file as the \code{prof_input} argument to
\code{profvis()}.
}
\examples{
# Only run these examples in interactive R sessions
if (interactive()) {

# Profile some code
profvis({
  dat <- data.frame(
    x = rnorm(5e4),
    y = rnorm(5e4)
  )

  plot(x ~ y, data = dat)
  m <- lm(x ~ y, data = dat)
  abline(m, col = "red")
})


# Save a profile to an HTML file
p <- profvis({
  dat <- data.frame(
    x = rnorm(5e4),
    y = rnorm(5e4)
  )

  plot(x ~ y, data = dat)
  m <- lm(x ~ y, data = dat)
  abline(m, col = "red")
})
htmlwidgets::saveWidget(p, "profile.html")

# Can open in browser from R
browseURL("profile.html")

}
}
\seealso{
\code{\link[=print.profvis]{print.profvis()}} for printing options.

\code{\link[=Rprof]{Rprof()}} for more information about how the profiling
data is collected.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/profvis-package.R
\docType{package}
\name{profvis-package}
\alias{profvis-package}
\title{profvis: Interactive Visualizations for Profiling R Code}
\description{
Interactive visualizations for profiling R code.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://profvis.r-lib.org}
  \item \url{https://github.com/r-lib/profvis}
  \item Report bugs at \url{https://github.com/r-lib/profvis/issues}
}

}
\author{
\strong{Maintainer}: Hadley Wickham \email{hadley@posit.co}

Authors:
\itemize{
  \item Winston Chang
  \item Javier Luraschi
  \item Timothy Mastny
}

Other contributors:
\itemize{
  \item Posit Software, PBC [copyright holder, funder]
  \item  jQuery Foundation (jQuery library) [copyright holder]
  \item  jQuery contributors (jQuery library; authors listed in inst/htmlwidgets/lib/jquery/AUTHORS.txt) [contributor, copyright holder]
  \item Mike Bostock (D3 library) [contributor, copyright holder]
  \item  D3 contributors (D3 library) [contributor]
  \item Ivan Sagalaev (highlight.js library) [contributor, copyright holder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parse.R
\name{parse_rprof}
\alias{parse_rprof}
\title{Parse Rprof output file for use with profvis}
\usage{
parse_rprof(path = "Rprof.out", expr_source = NULL)
}
\arguments{
\item{path}{Path to the \code{\link[=Rprof]{Rprof()}} output file.}

\item{expr_source}{If any source refs in the profiling output have an empty
filename, that means they refer to code executed at the R console. This
code can be captured and passed (as a string) as the \code{expr_source}
argument.}
}
\description{
Parse Rprof output file for use with profvis
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shiny_module.R
\name{profvis_ui}
\alias{profvis_ui}
\alias{profvis_server}
\title{profvis UI for Shiny Apps}
\usage{
profvis_ui(id)

profvis_server(input, output, session, dir = ".")
}
\arguments{
\item{id}{Output id from \code{profvis_server}.}

\item{input, output, session}{Arguments provided by
\code{\link[shiny:callModule]{shiny::callModule()}}.}

\item{dir}{Output directory to save Rprof files.}
}
\description{
Use this Shiny module to inject profvis controls into your Shiny app. The
profvis Shiny module injects UI that can be used to start and stop profiling,
and either view the results in the profvis UI or download the raw .Rprof
data. It is highly recommended that this be used for testing and debugging
only, and not included in production apps!
}
\details{
The usual way to use profvis with Shiny is to simply call
\code{profvis(shiny::runApp())}, but this may not always be possible or desirable:
first, if you only want to profile a particular interaction in the Shiny app
and not capture all the calculations involved in starting up the app and
getting it into the correct state; and second, if you're trying to profile an
application that's been deployed to a server.

For more details on how to invoke Shiny modules, see \href{https://shiny.rstudio.com/articles/modules.html}{this article}.
}
\examples{
# In order to avoid "Hit <Return> to see next plot" prompts,
# run this example with `example(profvis_ui, ask=FALSE)`

if(interactive()) {
  library(shiny)
  shinyApp(
    fluidPage(
      plotOutput("plot"),
      actionButton("new", "New plot"),
      profvis_ui("profiler")
    ),
    function(input, output, session) {
      callModule(profvis_server, "profiler")

      output$plot <- renderPlot({
        input$new
        boxplot(mpg ~ cyl, data = mtcars)
      })
    }
  )
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/profvis.R
\name{print.profvis}
\alias{print.profvis}
\title{Print a profvis object}
\usage{
\method{print}{profvis}(x, ..., width = NULL, height = NULL, split = NULL, aggregate = NULL)
}
\arguments{
\item{x}{The object to print.}

\item{...}{Further arguments to passed on to other print methods.}

\item{width}{Width of the htmlwidget.}

\item{height}{Height of the htmlwidget}

\item{split}{Orientation of the split bar: either \code{"h"} (the default) for
horizontal or \code{"v"} for vertical.}

\item{aggregate}{If \code{TRUE}, the profiled stacks are aggregated by
name. This makes it easier to see the big picture. Set your own
global default for this argument with \code{options(profvis.aggregate = )}.}
}
\description{
Print a profvis object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/profvis.R
\name{profvisOutput}
\alias{profvisOutput}
\alias{renderProfvis}
\title{Widget output and renders functions for use in Shiny}
\usage{
profvisOutput(outputId, width = "100\%", height = "600px")

renderProfvis(expr, env = parent.frame(), quoted = FALSE)
}
\arguments{
\item{outputId}{Output variable for profile visualization.}

\item{width}{Width of the htmlwidget.}

\item{height}{Height of the htmlwidget}

\item{expr}{An expression that returns a profvis object.}

\item{env}{The environment in which to evaluate \code{expr}.}

\item{quoted}{Is \code{expr} a quoted expression (with \code{\link[=quote]{quote()}})?}
}
\description{
Widget output and renders functions for use in Shiny
}
